"""Models for AI."""
from abc import ABC, ABCMeta, abstractmethod, abstractproperty
from dataclasses import dataclass
from typing import Optional, Iterator, List


@dataclass
class ModelMetrics:
    """Metrics for the model."""

    init_latency_ms: float


@dataclass
class CompletionMetrics:
    """Metrics related to a single completion."""

    """Number of tokens in the prompt for this completion."""
    prompt_tokens: int

    """Time taken for prompt evaluation."""
    prompt_eval_latency_ms: float

    """Number of tokens in this completion."""
    completion_tokens: int  # number of tokens in completions

    """Number of runs required to generate above tokens. Each run can generate
    more than one token.

    1 for streaming response. N for completed response.
    """
    completion_runs: int

    """Time taken for this completion."""
    completion_latency_ms: float


@dataclass
class CompletionResponse:
    """Response from a generative ai model."""

    text: str
    finish_reason: Optional[str]  # stop, length, none
    metrics: CompletionMetrics


class CompletionModel(ABC, metaclass=ABCMeta):
    """A generative AI model."""

    @abstractproperty
    def metrics(self) -> ModelMetrics:
        """Get metrics for the model."""
        raise NotImplementedError

    @abstractmethod
    def load(self, text: str):
        """Load the model with a warm up system prompt."""
        raise NotImplementedError

    @abstractmethod
    def complete(self, text: str) -> Iterator[CompletionResponse]:
        """Create a completion for given text."""
        raise NotImplementedError

    @abstractmethod
    def count_tokens(self, text: str) -> int:
        """Count tokens for the given text."""
        raise NotImplementedError


class EmbeddingModel(ABC, metaclass=ABCMeta):
    """An embedding AI model."""

    @property
    @abstractmethod
    def dimensions(self) -> int:
        """Count of dimensions supported by the model."""
        raise NotImplementedError

    @abstractmethod
    def embed(self, text: str) -> List[float]:
        """Create an embedding for given text."""
        raise NotImplementedError


def combine_metrics(usage_series: List[CompletionMetrics]) -> CompletionMetrics:
    """Join a series of completion metrics into one."""
    response_latency = 0
    response_tokens = 0
    for u in usage_series:
        response_latency += u.completion_latency_ms
        response_tokens += u.completion_tokens
    return CompletionMetrics(
        prompt_tokens=usage_series[0].prompt_tokens,
        prompt_eval_latency_ms=usage_series[0].prompt_eval_latency_ms,
        completion_tokens=response_tokens,
        completion_runs=len(usage_series),
        completion_latency_ms=response_latency,
    )



"""Create a abstract class for chat prompts."""
import os
from dataclasses import dataclass, field
from functools import lru_cache
from string import Template
from typing import Dict, List, Literal

import yaml

from arey.platform.assets import get_asset_dir
from arey.model import AreyError

SYSTEM_TOKENS = set(["message_text", "chat_history", "user_query"])


@dataclass
class Prompt:
    """An extensible prompt with dynamic template provided in a YML file.

    Tokens in the file are replaced at runtime.
    """

    name: str
    system_tokens: List[str] = field(default_factory=list)
    custom_tokens: Dict[str, str] = field(default_factory=dict)
    stop_words: List[str] = field(default_factory=list)
    prompts: Dict[str, str] = field(default_factory=dict)  # task: prompt
    message_formats: Dict[str, str] = field(default_factory=dict)  # role: format

    @classmethod
    def create(cls, yml: str) -> "Prompt":
        """Create a prompt from yml file."""
        content = yaml.safe_load(yml) or {}
        name = content.get("name", "")
        if not name:
            raise AreyError(
                "template", "`name` element in the prompt template is " "required."
            )

        tokens = content.get("tokens", {})
        system_tokens = tokens.get("system", [])
        custom_tokens = tokens.get("custom", {})

        stop_words = content.get("stop_words", [])

        prompts = content.get("prompts", {})
        if not prompts.get("chat") or not prompts.get("task"):
            raise AreyError(
                "template",
                "`prompts` element in the prompt template is required."
                " It must define `chat` and `task` prompt formats.",
            )

        roles = content.get("roles", {})
        if (
            not roles
            or not roles.get("ai")
            or not roles.get("user")
            or not roles.get("system")
        ):
            raise AreyError(
                "template", "`roles` element in the prompt template is required."
            )
        message_formats = {
            "ai": roles.get("ai").get("message"),
            "user": roles.get("user").get("message"),
            "system": roles.get("system").get("message"),
        }

        return cls(
            name, system_tokens, custom_tokens, stop_words, prompts, message_formats
        )

    @classmethod
    def create_overrides(cls, yml: str) -> "Prompt":
        """Create an override prompt. They are merged with the base prompt."""
        content = yaml.safe_load(yml) or {}
        name = content.get("name", "")
        if not name:
            raise AreyError(
                "template", "`name` element in the prompt template is required."
            )

        prompt_type = content.get("type", "")
        if not prompt_type:
            raise AreyError(
                "template", "`type` element in the prompt template is required."
            )

        # Overrides are only supported for custom_tokens currently
        tokens = content.get("tokens", {})
        custom_tokens = tokens.get("custom", {})

        return cls(name, custom_tokens=custom_tokens)

    def get(self, task: Literal["chat", "task"], context: Dict[str, str]) -> str:
        """Get a prompt with tokens resolved from the context."""
        merged_context = {**context, **self.custom_tokens}
        return Template(self.prompts[task]).substitute(merged_context)

    def get_message(
        self,
        role: Literal["ai", "user", "system"],
        text: str,
        token_overrides: Dict[str, str] = {},
    ) -> str:
        """Get a chat message for given role and text."""
        merged_context = {"message_text": text} | self.custom_tokens | token_overrides
        return Template(self.message_formats[role]).substitute(merged_context)


@lru_cache(maxsize=1)
def _get_oob_prompts() -> Dict[str, Prompt]:
    # get list of yml files in arey/prompts
    result: Dict[str, Prompt] = {}
    dir_path = get_asset_dir("prompts")
    for file_path in os.listdir(dir_path):
        if not file_path.endswith(".yml"):
            continue

        try:
            with open(os.path.join(dir_path, file_path), "r") as f:
                prompt = Prompt.create(f.read())
                result[prompt.name] = prompt
        except Exception as err:
            print(f"Error parsing: {file_path}. Error: {err.args[0]}")
            raise

    return result


def has_prompt(template_name: str) -> bool:
    """Check if prompt template is available."""
    oob_prompts = getattr(has_prompt, "oob_prompts", set())
    if not oob_prompts:
        oob_prompts = set(_get_oob_prompts().keys())
        setattr(has_prompt, "oob_prompts", oob_prompts)
    return template_name in oob_prompts


def get_prompt(template_name: str) -> Prompt:
    """Create an out-of-box prompt from template name."""
    oob_prompts = _get_oob_prompts()
    return oob_prompts[template_name]


def get_prompt_overrides(prompt_file_path: str) -> Prompt:
    """Create a prompt with overrides from a file."""
    try:
        with open(prompt_file_path, "r") as f:
            return Prompt.create_overrides(f.read())
    except Exception as err:
        print(f"Error parsing: {prompt_file_path}. Error: {err.args[0]}")
        raise



"""Arey app cli entrypoint."""
#!/usr/bin/env python
import click
import signal
import datetime
from functools import wraps
from typing import Callable, Iterable, Optional

from rich.console import Console, Group
from rich.live import Live
from rich.markdown import Markdown
from rich.padding import Padding
from rich.spinner import Spinner
from rich.text import Text

from watchfiles import watch

from arey.ai import CompletionMetrics
from arey.model import AreyError
from arey.platform.console import SignalContextManager, get_console


def _generate_response(
    console: Console,
    run: Callable[[], Iterable[str]],
    get_metrics: Callable[[], Optional[CompletionMetrics]],
) -> None:
    stop_completion = False

    def stop_completion_handler(signal, frame):
        nonlocal stop_completion
        stop_completion = True

    spinner = Spinner(
        text="[message_footer]Generating...", name="dots", style="message_footer"
    )
    text = Text()
    output = Group(
        Padding(text, pad=(0, 0, 2, 0)),
        spinner,
    )

    with SignalContextManager(signal.SIGINT, stop_completion_handler):
        with Live(
            output,
            console=console,
            transient=True,
        ):
            for response in run():
                if stop_completion:
                    break
                text.append(response)

    console.print(Markdown(text.plain))

    console.print()
    metrics = get_metrics()
    footer = "◼ Canceled." if stop_completion else "◼ Completed."
    if metrics:
        tokens_per_sec = (
            metrics.completion_tokens * 1000 / metrics.completion_latency_ms
        )
        footer += (
            f" {metrics.prompt_eval_latency_ms / 1000:.2f}s to first token."
            f" {metrics.completion_latency_ms / 1000:.2f}s total."
            f" {tokens_per_sec:.2f} tokens/s."
            f" {metrics.completion_tokens} tokens."
            f" {metrics.prompt_tokens} prompt tokens."
        )

    console.print()
    console.print(footer, style="message_footer")


def _print_logs(console: Console, verbose: bool, logs: Optional[str]) -> None:
    if not verbose or not logs:
        return
    console.print()
    console.print(logs)
    console.print()


def error_handler(func):
    """Global error handler for Arey."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except AreyError as e:
            console = get_console()
            help_text: str = (
                "Ouch! This appears to be a bug. Do you mind "
                "connecting with us at "
                "<https://github.com/codito/arey/issues/new>"
            )

            match e.category:
                case "template":
                    help_text = "A template seems misconfigured. Check out the docs."
                case "config":
                    help_text = "Config file seems misconfigured. Check out the docs."

            error_text = Group(
                Markdown(f"ERROR: {e.args[0]}", style="error"),
                Text(),
                Markdown(help_text),
            )
            console.print(error_text)

    return wrapper


@click.command("ask")
@click.argument("instruction", nargs=-1)
@click.option("-o", "--overrides-file", type=click.File())
@click.pass_context
@error_handler
def task(ctx: click.Context, instruction: str, overrides_file: str) -> int:
    """Run an instruction and generate response."""
    from arey.task import create_task, run

    verbose = ctx.obj["VERBOSE"]

    console = get_console()
    console.print()
    console.print("Welcome to arey task!")
    console.print()

    with console.status("[message_footer]Loading model..."):
        task, model_metrics = create_task(overrides_file)
        footer = f"✓ Model loaded. {model_metrics.init_latency_ms / 1000:.2f}s."
        console.print(footer, style="message_footer")
        console.print()

    _generate_response(
        console,
        lambda: run(task, instruction),
        lambda: (task.result and task.result.metrics),
    )

    _print_logs(console, verbose, task.result and task.result.logs)
    return 0


@click.command("chat")
@click.pass_context
@error_handler
def chat(ctx: click.Context) -> int:
    """Chat with an AI model."""
    from arey.chat import create_chat, get_completion_metrics, stream_response

    verbose = ctx.obj["VERBOSE"]

    console = get_console()
    console.print(("Welcome to arey chat!\nType 'q' to exit."))
    console.print()

    with console.status("[message_footer]Loading model..."):
        chat, model_metrics = create_chat()
        footer = f"✓ Model loaded. {model_metrics.init_latency_ms / 1000:.2f}s."
        console.print(footer, style="message_footer")
        console.print()

    console.print("How can I help you today?")
    while True:
        # Get input from user
        # Workaround for https://github.com/Textualize/rich/issues/2293
        with console.capture() as capture:
            console.print("> ", style="message_prompt", end="")
        prompt_str = capture.get()
        try:
            user_input = input(prompt_str)
        except KeyboardInterrupt:
            console.print()
            continue
        except EOFError:
            console.print("\nBye!")
            break

        if user_input in ["q", "quit"]:
            console.print("Bye!")
            break

        console.print()
        _generate_response(
            console,
            lambda: stream_response(chat, user_input),
            lambda: get_completion_metrics(chat),
        )

        _print_logs(
            console,
            verbose,
            (
                chat.messages[-1].context.logs
                if chat.messages and chat.messages[-1].context
                else ""
            ),
        )

    return 0


@click.command("play")
@click.argument("file", required=False)
@click.option(
    "--no-watch",
    is_flag=True,
    default=False,
    help="Watch the play file and regenerate response on save.",
)
@click.pass_context
@error_handler
def play(ctx: click.Context, file: str, no_watch: bool) -> int:
    """Watch FILE for model, prompt and generate response on edit.

    If FILE is not provided, a temporary file is created for edit.
    """
    from arey.play import get_play_file, get_play_response, load_play_model

    verbose = ctx.obj["VERBOSE"]

    console = get_console()
    console.print()
    console.print(
        "Welcome to arey play! Edit the play file below in your favorite editor "
        "and I'll generate a response for you. Use `Ctrl+C` to abort play session."
    )
    console.print()

    play_file = get_play_file(file)

    def run_play_file():
        file_path = file or play_file.file_path
        current_date = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        console.print()
        console.rule(current_date)
        play_file_mod = get_play_file(file_path)
        with console.status("[message_footer]Loading model..."):
            model_metrics = load_play_model(play_file_mod)
            footer = f"✓ Model loaded. {model_metrics.init_latency_ms / 1000:.2f}s."
            console.print(footer, style="message_footer")
            console.print()

        _generate_response(
            console,
            lambda: get_play_response(play_file_mod),
            lambda: (play_file_mod.result and play_file_mod.result.metrics),
        )

        _print_logs(console, verbose, play_file.result and play_file.result.logs)

    if no_watch:
        run_play_file()
        return 0

    console.print(f"Watching `{play_file.file_path}` for changes...")
    for _ in watch(play_file.file_path):
        run_play_file()
        console.print()
        console.print(f"Watching `{play_file.file_path}` for changes...")
    return 0


@click.group()
@click.option("-v", "--verbose", default=False, help="Show verbose logs.")
@click.pass_context
def main(ctx: click.Context, verbose: bool):
    """Arey - a simple large language model app."""
    ctx.ensure_object(dict)
    ctx.obj["VERBOSE"] = verbose
    pass


main.add_command(task)
main.add_command(chat)
main.add_command(play)

if __name__ == "__main__":
    main()



"""Configuration for arey."""
import os
from dataclasses import dataclass, field
from typing import Dict, Optional, TypedDict, Tuple, Union, cast

import yaml

from arey.model import AreyError
from arey.platform.assets import get_config_dir, get_default_config


@dataclass
class ModelConfig:
    """Configuration for a given model."""

    path: str
    template: str
    type: Optional[str] = "llama2"


class ProfileConfig(TypedDict):
    """Configuration for a given profile."""

    temperature: float
    repeat_penalty: float
    top_k: int
    top_p: float


@dataclass
class ChatConfig:
    """Configuration for chat mode."""

    model_name: str
    model: ModelConfig
    profile: ProfileConfig
    settings: Dict = field(default_factory=dict)


@dataclass
class TaskConfig:
    """Configuration for task mode."""

    model_name: str
    model: ModelConfig
    profile: ProfileConfig
    settings: Dict = field(default_factory=dict)


@dataclass
class Config:
    """Arey Configuration."""

    models: Dict[str, ModelConfig]
    profiles: Dict[str, ProfileConfig]
    chat: ChatConfig
    task: TaskConfig

    @classmethod
    def from_dict(cls, config: dict):
        """Create a configuration from dictionary."""
        from arey.prompt import has_prompt

        models = {
            key: ModelConfig(val["path"], val["template"], val.get("type", "llama2"))
            for key, val in config.get("models", {}).items()
        }
        profiles = {
            key: ProfileConfig(**val) for key, val in config.get("profiles", {}).items()
        }
        default_profile: ProfileConfig = {
            "temperature": 0.7,
            "repeat_penalty": 1.176,
            "top_k": 40,
            "top_p": 0.1,
        }

        if "chat" not in config or "task" not in config:
            raise AreyError(
                "config", "`chat` and `task` sections are not available in config file."
            )

        def _get_config(key: str) -> Union[ChatConfig, TaskConfig]:
            model_name = config[key].get("model", None)
            if not model_name or model_name not in models:
                raise AreyError(
                    "config", f"Section '{key}' must have valid `model` entry."
                )
            model = models[model_name]
            model.path = os.path.expanduser(model.path) if model.path else model.path
            if not os.path.exists(model.path):
                raise AreyError(
                    "config", f"Model '{model_name}' has invalid path: {model.path}."
                )
            if not has_prompt(model.template):
                raise AreyError(
                    "config",
                    f"Model '{model_name}' has invalid template: {model.template}.",
                )

            profile = profiles.get(
                config[key].get("profile", "default"), default_profile
            )
            settings = config[key]["settings"] if "settings" in config[key] else {}
            if key == "chat":
                return ChatConfig(model_name, model, profile, settings)
            return TaskConfig(model_name, model, profile, settings)

        chat = _get_config("chat")
        task = _get_config("task")
        return cls(models, profiles, cast(ChatConfig, chat), cast(TaskConfig, task))


def create_or_get_config_file() -> Tuple[bool, str]:
    """Get config file path if exists, create a default otherwise."""
    config_file = os.path.join(get_config_dir(), "arey.yml")
    if os.path.exists(config_file):
        return (True, config_file)

    # Create a default config, ask user to update the model
    with open(config_file, "w", encoding="utf-8") as f:
        f.write(get_default_config())
    return (False, config_file)


def get_config() -> Config:
    """Get the app configuration if available."""
    config = getattr(get_config, "config", None)
    if config:
        return config

    _, config_file = create_or_get_config_file()
    with open(config_file, "r") as f:
        config = Config.from_dict(yaml.safe_load(f) or {})
        setattr(get_config, "config", config)
        return config



"""Services for the chat command."""
from typing import Iterator, Optional, Tuple

from arey.ai import CompletionMetrics, ModelMetrics, combine_metrics
from arey.config import get_config
from arey.model import Chat, Message, MessageContext, SenderType
from arey.platform.console import capture_stderr
from arey.platform.llama import LlamaBaseModel
from arey.prompt import Prompt, get_prompt

config = get_config()
model_path = config.chat.model.path
prompt_template = config.chat.model.template
model_settings = config.chat.settings
completion_settings = config.chat.profile

model: LlamaBaseModel = LlamaBaseModel(
    model_path=model_path, model_settings=model_settings
)
prompt_model = get_prompt(prompt_template)


def _get_max_tokens(model: LlamaBaseModel, prompt_model: Prompt, text: str) -> int:
    context_size = model.context_size
    prompt_tokens_without_history = model.count_tokens(
        prompt_model.get("chat", {"user_query": text, "chat_history": ""})
    )
    buffer = 200

    return context_size - prompt_tokens_without_history - buffer


def create_chat() -> Tuple[Chat, ModelMetrics]:
    """Create a new chat session."""
    system_prompt = prompt_model.get_message("system", "")
    with capture_stderr() as stderr:
        model.load(system_prompt)
    chat = Chat()
    chat.context.metrics = model.metrics
    chat.context.logs = stderr.getvalue()
    return chat, model.metrics


def get_history(
    model: LlamaBaseModel, chat: Chat, prompt_model: Prompt, max_tokens: int
) -> str:
    """Get the messages for a chat."""
    messages = []
    token_count = 0
    for message in reversed(chat.messages):
        role = (
            "ai"
            if message.sender == SenderType.AI
            else "user"
            if message.sender == SenderType.USER
            else "system"
        )
        formatted_message = prompt_model.get_message(role, message.text)
        message_tokens = model.count_tokens(formatted_message)
        messages.insert(0, formatted_message)

        token_count += message_tokens
        if message.sender == SenderType.USER and token_count >= max_tokens:
            break

    return "".join(messages)


def create_response(chat: Chat, message: str) -> str:
    """Create a chat response."""
    response = ""
    for chunk in stream_response(chat, message):
        response += chunk

    return response


def stream_response(chat: Chat, message: str) -> Iterator[str]:
    """Stream a chat response."""
    max_tokens = _get_max_tokens(model, prompt_model, message)
    context = {
        "user_query": message,
        "chat_history": get_history(model, chat, prompt_model, max_tokens),
    }
    prompt = prompt_model.get("chat", context)

    user_msg = Message(message, 0, SenderType.USER, None)
    chat.messages.append(user_msg)

    ai_msg_text = ""
    usage_series = []
    finish_reason = ""
    with capture_stderr() as stderr:
        for chunk in model.complete(prompt, {"stop": prompt_model.stop_words}):
            ai_msg_text += chunk.text
            finish_reason = chunk.finish_reason
            usage_series.append(chunk.metrics)
            yield chunk.text

    msg_context = MessageContext(
        prompt=prompt,
        finish_reason=finish_reason,
        metrics=combine_metrics(usage_series),
        logs=stderr.getvalue(),
    )
    ai_msg = Message(ai_msg_text, 0, SenderType.AI, msg_context)
    chat.messages.append(ai_msg)


def get_completion_metrics(chat: Chat) -> Optional[CompletionMetrics]:
    """Get completion metrics for the chat."""
    msg = next(
        filter(lambda m: m.sender == SenderType.AI, reversed(chat.messages)), None
    )
    return msg.context.metrics if msg and msg.context else None



"""Services for the task command.

A task is stateless execution of a single instruction.
"""
import os
from dataclasses import dataclass, field
from typing import Iterator, Optional, Tuple

from arey.ai import CompletionMetrics, ModelMetrics, combine_metrics
from arey.config import get_config
from arey.platform.console import capture_stderr
from arey.platform.llama import LlamaBaseModel
from arey.prompt import get_prompt, get_prompt_overrides

config = get_config()
model_path = config.chat.model.path
prompt_template = config.chat.model.template
model_settings = config.chat.settings
completion_settings = config.chat.profile

model: LlamaBaseModel = LlamaBaseModel(
    model_path=model_path, model_settings=model_settings
)
prompt_model = get_prompt(prompt_template)


@dataclass
class TaskResult:
    """Result of a task execution."""

    response: str
    metrics: CompletionMetrics
    finish_reason: Optional[str]
    logs: Optional[str]


@dataclass
class Task:
    """A task is a stateless script invocation with a prompt."""

    prompt_overrides: dict = field(default_factory=dict)
    result: Optional[TaskResult] = None


def create_task(prompt_file: Optional[str]) -> Tuple[Task, ModelMetrics]:
    """Create a task with given prompt file."""
    token_overrides = (
        get_prompt_overrides(prompt_file).custom_tokens
        if prompt_file and os.path.exists(prompt_file)
        else {}
    )
    system_prompt = prompt_model.get_message("system", "", token_overrides)
    with capture_stderr():
        model.load(system_prompt)
    task = Task()
    return task, model.metrics


def run(task: Task, user_input: str) -> Iterator[str]:
    """Run a task with user query."""
    context = {
        "user_query": user_input,
        "chat_history": "",
    }
    prompt = prompt_model.get("task", context)

    ai_msg_text = ""
    usage_series = []
    finish_reason = ""
    with capture_stderr() as stderr:
        for chunk in model.complete(prompt, {"stop": prompt_model.stop_words}):
            ai_msg_text += chunk.text
            finish_reason = chunk.finish_reason
            usage_series.append(chunk.metrics)
            yield chunk.text

    task.result = TaskResult(
        response=ai_msg_text,
        metrics=combine_metrics(usage_series),
        finish_reason=finish_reason,
        logs=stderr.getvalue(),
    )



"""Abstraction for various data and config assets in Arey."""
import os


DEFAULT_DATA_DIR = os.path.expanduser(
    "~/.local/share/arey" if os.name == "posix" else "~/.arey"
)
DEFAULT_CONFIG_DIR = os.path.expanduser(
    "~/.config/arey" if os.name == "posix" else "~/.arey"
)


def _make_dir(path: str) -> None:
    if not os.path.exists(path):
        os.mkdir(path)


def get_asset_dir(suffix: str = "") -> str:
    """Get path of the assets directory.

    params:
        suffix (str): suffix directory name to append
    """
    dir_path = os.path.dirname(os.path.realpath(__file__))
    return os.path.abspath(os.path.join(dir_path, "..", "data", suffix))


def get_asset_path(asset_name: str) -> str:
    """Get the full path to included asset.

    params:
        asset_name (str): relative path to the asset. E.g., prompts/alpaca.yml
    """
    dir_path = get_asset_dir()
    file_path = os.path.join(dir_path, asset_name)
    if not os.path.exists(file_path):
        raise FileNotFoundError(
            f"Asset file {asset_name} not found. Resolved path: {file_path}"
        )
    return file_path


def get_config_dir():
    """Get arey config dir."""
    base_dir = os.environ.get("XDG_CONFIG_HOME")
    config_dir = os.path.join(base_dir, "arey") if base_dir else DEFAULT_CONFIG_DIR
    _make_dir(config_dir)
    return config_dir


def get_default_config() -> str:
    """Get default configuration template."""
    config_file = get_asset_path("config.yml")
    with open(config_file, "r", encoding="utf-8") as f:
        return f.read()



"""Core data structures for the arey app."""
from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional, Union, Literal

from arey.ai import CompletionMetrics, ModelMetrics


class SenderType(Enum):
    """Chat message sender."""

    SYSTEM = 1
    AI = 2
    USER = 3


@dataclass
class MessageContext:
    """Context associated with a single chat message."""

    prompt: str
    finish_reason: Optional[str]
    metrics: CompletionMetrics
    logs: str = ""


@dataclass
class Message:
    """A chat message."""

    text: str
    timestamp: int  # unix timestamp
    sender: SenderType
    context: MessageContext


@dataclass
class ChatContext:
    """Context associated with a chat."""

    metrics: Optional[ModelMetrics] = None
    logs: str = ""


@dataclass
class Chat:
    """A chat conversation between human and AI model."""

    messages: List[Message] = field(default_factory=list)
    context: ChatContext = field(default_factory=ChatContext)


class AreyError(Exception):
    """Error in Arey execution."""

    category: Union[Literal["config"], Literal["template"], Literal["system"]]
    message: str

    def __init__(
        self,
        category: Union[Literal["config"], Literal["template"], Literal["system"]],
        message: str,
    ):
        """Create an instance of AreyError with category and message."""
        self.category = category
        self.message = message
        super().__init__(message)



"""Services for the play command.

Play reads the model details from a markdown file, and the prompt. It will watch
the file for any changes and run a completion for the prompt.
"""
from dataclasses import dataclass
import os
import tempfile
import frontmatter
from functools import lru_cache
from arey.ai import CompletionMetrics, ModelMetrics, combine_metrics
from arey.config import ModelConfig, get_config
from arey.model import AreyError
from arey.platform.assets import get_asset_path
from arey.platform.llama import LlamaBaseModel
from arey.platform.console import capture_stderr
from typing import Dict, Optional, Iterator


config = get_config()


@dataclass
class PlayResult:
    """Result of a task execution."""

    response: str
    metrics: CompletionMetrics
    finish_reason: Optional[str]
    logs: Optional[str]


@dataclass
class PlayFile:
    """A play file with model settings and a prompt."""

    file_path: str

    model_config: ModelConfig
    model_settings: Dict

    prompt: str
    completion_profile: Dict

    model: Optional[LlamaBaseModel] = None
    result: Optional[PlayResult] = None


@lru_cache(maxsize=1)
def _get_default_play_file() -> str:
    """Get the default play file template."""
    file_path = get_asset_path("play.md")
    with open(file_path, "r", encoding="utf-8") as f:
        return f.read()


def _create_file_if_not_exists(play_file_path: str) -> str:
    if play_file_path and os.path.exists(play_file_path):
        return play_file_path

    with tempfile.NamedTemporaryFile(
        mode="w+b", prefix="arey_play", suffix=".md", delete=False
    ) as file:
        file.write(_get_default_play_file().encode())
        return file.name


def get_play_file(file_path: str) -> PlayFile:
    """Evaluate the provided play file and generate a response."""
    play_file_path = _create_file_if_not_exists(file_path)
    with open(play_file_path, "r", encoding="utf-8") as f:
        play_file = frontmatter.load(f)

    # FIXME validate settings
    model_config = config.models[play_file.metadata["model"]]
    model_settings = play_file.metadata["settings"]
    return PlayFile(
        file_path=play_file_path,
        model_config=model_config,
        model_settings=model_settings,
        prompt=play_file.content,
        completion_profile={},
    )


def load_play_model(play_file: PlayFile) -> ModelMetrics:
    """Load a model from play file."""
    model_config = play_file.model_config
    model_settings = play_file.model_settings
    with capture_stderr():
        model: LlamaBaseModel = LlamaBaseModel(
            model_path=model_config.path, model_settings=model_settings
        )
        model.load("")
        play_file.model = model
        return model.metrics


def get_play_response(play_file: PlayFile) -> Iterator[str]:
    """Run a task with user query."""
    model = play_file.model
    if not model:
        raise AreyError("system", "Model is empty.")

    completion_settings = play_file.completion_profile
    prompt = play_file.prompt
    ai_msg_text = ""
    usage_series = []
    finish_reason = ""
    with capture_stderr() as stderr:
        for chunk in model.complete(prompt, completion_settings):
            ai_msg_text += chunk.text
            finish_reason = chunk.finish_reason
            usage_series.append(chunk.metrics)
            yield chunk.text

    play_file.result = PlayResult(
        response=ai_msg_text,
        metrics=combine_metrics(usage_series),
        finish_reason=finish_reason,
        logs=stderr.getvalue(),
    )



"""Llama.cpp based models."""
import dataclasses
import os
import time
import multiprocessing
from typing import Iterator, cast

import llama_cpp

from arey.ai import CompletionMetrics, CompletionModel, CompletionResponse, ModelMetrics
from arey.model import AreyError


@dataclasses.dataclass
class LlamaSettings:
    """Core model settings."""

    n_threads: int = max(multiprocessing.cpu_count() // 2, 1)
    n_ctx: int = 4096
    n_batch: int = 512
    n_gpu_layers: int = 0
    verbose: bool = True


class LlamaBaseModel(CompletionModel):
    """Base local completion model.

    Wraps over the llama-cpp library.
    """

    context_size: int = 4096
    _model_settings: LlamaSettings
    _metrics: ModelMetrics

    def __init__(self, model_path: str, model_settings: dict = {}) -> None:
        """Create an instance of local completion model."""
        self._llm = None
        self._model_path = model_path
        self._model_settings = LlamaSettings(**model_settings)
        self.context_size = self._model_settings.n_ctx

    @property
    def metrics(self) -> ModelMetrics:
        """Get metrics for model initialization."""
        return self._metrics

    def _get_model(self):
        model_path = os.path.join(os.path.expanduser(self._model_path))
        if not os.path.exists(model_path):
            raise AreyError(
                "system",
                f"Invalid model path: {model_path}.",
            )
        if not self._llm:
            start_time = time.perf_counter()
            self._llm = llama_cpp.Llama(
                model_path=model_path,
                **dataclasses.asdict(self._model_settings),
            )
            # self._llm.set_cache(llama_cpp.LlamaCache(2 << 33))

            latency_ms = round((time.perf_counter() - start_time) * 1000, 2)
            self._metrics = ModelMetrics(init_latency_ms=latency_ms)
        return self._llm

    def load(self, text: str):
        """Load a model into memory."""
        model = self._get_model()
        model.eval(model.tokenize(text.encode("utf-8")))

    def complete(self, text: str, settings: dict = {}) -> Iterator[CompletionResponse]:
        """Get a completion for the given text and settings."""
        prev_time = time.perf_counter()
        model = self._get_model()
        output = cast(
            Iterator[llama_cpp.CompletionChunk],
            model.create_completion(
                prompt=text,
                max_tokens=-1,
                temperature=0.7,
                top_k=40,
                top_p=0.1,
                repeat_penalty=1.176,
                echo=False,
                **settings,
                stream=True,
            ),
        )

        prompt_token_count = self.count_tokens(text)
        prompt_eval_latency = -1
        for chunk in output:
            chunk_text = chunk["choices"][0]["text"]

            current_time = time.perf_counter()
            latency = current_time - prev_time
            prev_time = current_time
            if prompt_eval_latency == -1:
                prompt_eval_latency = round(latency * 1000, 2)

            token_count = self.count_tokens(chunk_text)
            yield CompletionResponse(
                text=chunk_text,
                finish_reason=chunk["choices"][0]["finish_reason"],
                metrics=CompletionMetrics(
                    prompt_token_count,
                    prompt_eval_latency,
                    token_count,
                    1,
                    round(latency * 1000, 2),
                ),
            )

    def count_tokens(self, text: str) -> int:
        """Get the token count for given text."""
        model = self._get_model()
        return len(model.tokenize(text.encode("utf-8")))



"""Console platform abstraction."""

import io
import signal
from contextlib import contextmanager, redirect_stderr
from functools import lru_cache
from io import StringIO
from typing import Generator

from rich.console import Console
from rich.theme import Theme
from wurlitzer import pipes

theme = Theme(
    {
        "header": "",
        "message_prompt": "bold blue",
        "message_user": "blue",
        "message_ai": "bright_white",
        "message_footer": "dim white",
        "error": "bold red",
    }
)


@lru_cache(maxsize=None)
def get_console() -> Console:
    """Get a console instance."""
    return Console(theme=theme)


class SignalContextManager:
    """Context manager for console signals."""

    def __init__(self, signal_num, handler):
        """Create a signal context instance."""
        self.signal_num = signal_num
        self.handler = handler
        self.original_handler = None

    def __enter__(self):
        """Context manager enter implementation."""
        self.original_handler = signal.signal(self.signal_num, self.handler)

    def __exit__(self, exc_type, exc_value, traceback):
        """Context manager exit implementation."""
        signal.signal(self.signal_num, self.original_handler)


@contextmanager
def capture_stderr() -> Generator[StringIO, None, None]:
    """Capture stderr for both python and c-functions."""
    stderr = io.StringIO()
    with redirect_stderr(stderr) as err, pipes(
        stdout=0, stderr=stderr, encoding="utf-8"
    ):
        yield err
